import hashlib
from lhub_integ.common import file_manager_client


def get_hash_custom(lhub_file_id, hashlib_object):
    """
    Takes a file and generates a hash based on the provided hashing object. The hashing object can either be one of
    hashlib's builtin functions or a custom object defined with hashlib.new(). (Note that there are other functions
    below to make it easier to get a specific type of hash.)

    Args:
        lhub_file_id (str):
            Name of a file created using file_manager_client

        hashlib_object (_hashlib.HASH object):
            Either a custom hash object created by hashlib.new() or one of the built-in functions:
                hashlib.md5()
                hashlib.sha1()
                hashlib.sha224()
                hashlib.sha256()
                hashlib.sha384()
                hashlib.sha512()

    Returns:
        (str): File hash string

    """
    file_to_hash = file_manager_client.openFileForReading(lhub_file_id, 'rb')
    for chunk in iter(lambda: file_to_hash.fd.read(4096), b""):
        hashlib_object.update(chunk)
    file_to_hash.close()
    return hashlib_object.hexdigest()


def get_hash_md5(lhub_file_id):
    return get_hash_custom(lhub_file_id, hashlib.md5())


def get_hash_sha1(lhub_file_id):
    return get_hash_custom(lhub_file_id, hashlib.sha1())


def get_hash_sha256(lhub_file_id):
    return get_hash_custom(lhub_file_id, hashlib.sha256())


def get_file_size(lhub_file_id):
    f = file_manager_client.openFileForReading(lhub_file_id)
    s = f.size()
    f.close()
    return s


def get_mime_type(lhub_file_id):
    return file_manager_client.get_mime_type(lhub_file_id)


def get_file_encoding(lhub_file_id):
    f = file_manager_client.openFileForReading(lhub_file_id)
    encoding_info = f.encoding()
    return encoding_info


def get_file_details(lhub_file_id):
    # compute the md5 hash
    md5_hash = get_hash_md5(lhub_file_id)
    # compute the sha1 hash
    sha1_hash = get_hash_sha1(lhub_file_id)
    # compute the sha256 hash
    sha256_hash = get_hash_sha256(lhub_file_id)
    # get the size of the file
    size = get_file_size(lhub_file_id)
    # guess its mime-type
    mime = get_mime_type(lhub_file_id)
    # guess its encoding
    encoding_info = get_file_encoding(lhub_file_id)
    result = {
        "hash_md5": md5_hash,
        "hash_sha1": sha1_hash,
        "hash_sha256": sha256_hash,
        "size": size,
        "content_type": mime,
        "encoding_info": encoding_info,
    }
    return result


def compute_file_hash(lhub_file_id):
    """
    Legacy method used to generate only an MD5 hash. This should no longer be used in new integrations. For md5 hashes,
    use get_hash_md5

    ToDo Remove this method once it is no longer used by any integrations
    """
    file_to_hash = file_manager_client.openFileForReading(lhub_file_id)
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: file_to_hash.fd.read(4096), b""):
        hash_md5.update(chunk)
    file_to_hash.close()
    return hash_md5.hexdigest()



import requests
from lhub_integ.common.helpers import format_success, format_error


# ToDo Duplicate of helpers.format_error
#   Confirm whether this is used anywhere else, switch to the other, and remove from here.
def error_message_json(message):
    return format_error(message)


def derive_result_json(response):
    if response.status_code == 401:
        result = format_error("Authentication failed. Please check the credentials and try again.")

    elif response.status_code == 403:
        result = format_error("Authorization failed. The credentials provided are valid but do not "
                              "have proper access level.")

    elif response.status_code == 404:
        result = format_error("Not found")

    elif response.status_code != requests.codes.ok:
        result = format_error("Unexpected status code %s: %s." % (response.status_code, response.reason))

    else:
        response_json = response.json()
        result = format_success(response_json)
    return result



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
the web url matching regex used by markdown
http://daringfireball.net/2010/07/improved_regex_for_matching_urls
https://gist.github.com/gruber/8891611
"""
URL_REGEX = r"""(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))"""



import time


def handle_request_throttling(wait_time_in_millis, req_func, max_retries=1):
    """
    This function handles throttling of request, it waits for specified time and retries.
    :param wait_time_in_millis:
    :param req_func: request function that returns response object. e.g lambda : requests.get("www.example.com")
    :param max_retries: default = 1,
    :return: response
    """
    assert max_retries >= 0, "Invalid max_retries argument: '{}', value must be non negative".format(max_retries)
    response = req_func()
    while response.status_code == 429 and max_retries:
        time.sleep(wait_time_in_millis / 1000.0)
        response = req_func()
        max_retries -= 1

    return response



"""
Wrap a custom Python script to enable it to work with the integration machinery

Loads ENTRYPOINT from `__ENTRYPOINT` and the column-argument mapping from `___XXXXX`

Usage:
This module is intended to be invoked by `PythonEnvironment`. See `test_shim.py` for a usage example

<set required environment variables> python -m lhub_integ.shim_exec -e <entrypoint>
"""
import json
import sys
import traceback

import click

from docstring_parser import parse

from lhub_integ import util, action, params
from lhub_integ.env import MappedColumnEnvVar
from lhub_integ.params import JinjaTemplatedStr
from lhub_integ.execution import ExecutionType


def _print_result_rows(result, lhub_id=None):
    # Print the results
    if result:
        if isinstance(result, (list, set)):
            util.print_each_result_in_list(result, lhub_id)
        else:
            util.print_result(json.dumps(result), lhub_id)


def run_integration(entrypoint_fn):
    type_mapper = params.get_input_converter(entrypoint_fn)

    annotations = util.type_annotations(entrypoint_fn)
    parsed = parse(entrypoint_fn.__doc__)
    # Build a mapping from the columns in the input data to the arguments in our integration
    argument_column_mapping = {}
    for input_argument in type_mapper:
        optional = False
        for meta in parsed.meta:
            if len(meta.args) == 2 and meta.args[0] == "optional" and meta.args[1] == input_argument:
                optional = meta.description.lower() != "false"

        input_value = MappedColumnEnvVar(input_argument, optional=optional).read()
        argument_column_mapping[input_argument] = input_value

    required_columns = {
        k: v
        for k, v in argument_column_mapping.items()
        if annotations.get(k) != JinjaTemplatedStr and v is not None
    }

    for row in sys.stdin.readlines():
        as_dict = json.loads(row)["row"]
        lhub_id = as_dict.get(util.LHUB_ID)
        missing_keys = required_columns.values() - as_dict.keys()
        for column in missing_keys:
            util.print_error(f"Column '{column}' is not present in the data")
        if missing_keys:
            continue
        try:

            # Load the arguments from the input
            def coerce(arg, column):
                try:
                    return type_mapper[arg](as_dict, column)
                except ValueError as ex:
                    util.print_error(
                        f"Invalid value for column {column}: [{ex}]", data=as_dict
                    )
                    raise EndOfLoop()

            arguments = {}
            for arg, column in argument_column_mapping.items():
                arguments[arg] = coerce(arg, column) if column is not None else column

            # Run the integration
            result = entrypoint_fn(**arguments)

            _print_result_rows(result, lhub_id)
        except EndOfLoop:
            pass
        except Exception:
            util.print_error(traceback.format_exc(), data=as_dict, original_lhub_id=lhub_id)


# Uncorrelated results. Used in actions that are `isDataSource`
def run_integration_once(entrypoint_fn):
    result = entrypoint_fn()
    _print_result_rows(result)


class EndOfLoop(Exception):
    pass


@click.command()
@click.option("--entrypoint", "-e", required=True)
@click.option("--compileall/--no-compileall", default=True)
def main(entrypoint, compileall):
    if compileall:
        errors, _ = util.import_workdir()
    else:
        errors, _ = util.import_file_of_entryPoint(entrypoint)
    if errors:
        util.hard_exit_from_instantiation(str(errors[0]))
    entrypoint_fn = action.all().get(entrypoint).function
    execution_type = action.all().get(entrypoint).execution_type
    assert entrypoint_fn is not None
    if execution_type == ExecutionType.PROCESS_ONCE:
        run_integration_once(entrypoint_fn)
    else:
        run_integration(entrypoint_fn)


if __name__ == "__main__":
    main()



import os
import socket
from urllib.parse import urlparse

import socks

# If socks_proxy is set in environment variable then,
# this script replaces the default socket to route through
# socks proxy

DEFAULT_SOCKET = socket.socket


def get_proxy():
    proxy = os.environ.get('socks_proxy')
    if proxy:
        parsed_url = urlparse(proxy)
        host = parsed_url.hostname
        port = parsed_url.port
        if host and port:
            return host, port
        return None
    else:
        return None


def override_socket():
    proxy_settings = get_proxy()
    if proxy_settings:
        host, port = proxy_settings
        socks.setdefaultproxy(socks.PROXY_TYPE_SOCKS5, host, port)
        socket.socket = socks.socksocket


def remove_socket_override():
    socket.socket = DEFAULT_SOCKET


override_socket()



import os
import json
import yaml


def safe_strip(value):
    if value:
        return value.strip()
    else:
        return value


# Strips whitespaces and attempts to parse integer from environment variable
# Or returns default value if no value is provided
# NOT SAFE against parsing failure for non-integers, will throw a ValueError
def __get_stripped_env_integer(env_var, default=0):
    assert (type(default) == int), "Default value must be an int"
    raw_value = os.environ.get(env_var)
    stripped_value = safe_strip(raw_value)
    if stripped_value:
        return int(stripped_value)
    else:
        return default


# Strips whitespaces and attempts to parse float from environment variable
# Or returns default value if no value is provided
# NOT SAFE against parsing failure for non-float or non integer, will throw a ValueError
def __get_stripped_env_float(env_var, default=0.0):
    if type(default) is int:
        default = float(default)
    else:
        assert (type(default) == float), "Default value must be a float"
    raw_value = os.environ.get(env_var)
    stripped_value = safe_strip(raw_value)
    if stripped_value:
        return float(stripped_value)
    else:
        return default


# Safe version of __get_stripped_env_integer
def _get_safe_stripped_env_integer(env_var, default=0):
    try:
        return __get_stripped_env_integer(env_var, default)
    except:
        return default


# Use in place of os.environ.get()
def _get_stripped_env_string(env_var, default=""):
    raw_value = os.environ.get(env_var)
    stripped_value = safe_strip(raw_value)
    return stripped_value if stripped_value else default


# Use in place of os.environ.get()
def _get_stripped_env_json(env_var, default={}):
    raw_value = os.environ.get(env_var)
    stripped_value = safe_strip(raw_value)
    return json.loads(stripped_value) if stripped_value else default


def _get_stripped_env_yaml(env_var, default=""):
    raw_value = os.environ.get(env_var)
    stripped_value = safe_strip(raw_value)
    return yaml.load(stripped_value) if stripped_value else default


def _get_json_from_string(raw_value, default={}):
    stripped_value = safe_strip(raw_value)
    return json.loads(stripped_value) if stripped_value else default


# Separates comma-separated input into a list of str
# with each element having trailing and leading whitespace removed
# Tolerant to list notation of square brackets
# Examples:
#   None         ->  []
#   '[1,2,3]'    ->  ['1', '2', '3']
#   'a,b,   c  ' ->  ['a', 'b', 'c']
def eval_to_stripped_list(input_str, delimiter=','):
    if not input_str:
        return []
    else:
        input_str_stripped = input_str.lstrip('[').rstrip(']')
        str_list = input_str_stripped.split(delimiter)
        return list(map(safe_strip, str_list))


def eval_to_stripped_list_or_none(input_str, delimiter=','):
    if not input_str:
        return None
    else:
        input_str_stripped = input_str.lstrip('[').rstrip(']')
        str_list = input_str_stripped.split(delimiter)
        return list(map(safe_strip, str_list))



"""
Validation Shim for custom integrations
"""
import inspect
from traceback import format_exc
from docstring_parser import parse

import click

from lhub_integ import action, util, connection_validator
from lhub_integ.env import MappedColumnEnvVar
from lhub_integ.params import ConnectionParam, ActionParam, ValidationError
from lhub_integ.util import (
    exit_with_instantiation_errors,
    print_successful_validation_result,
)
from lhub_integ.execution import ExecutionType


@click.command()
@click.option("--entrypoint", "-e", required=False)
@click.option(
    "--validate-connections/--no-validate-connections",
    "check_connections",
    default=False,
)
def main(entrypoint, check_connections):
    validation_errors = []
    try:
        errors, _ = util.import_workdir()
        if errors:
            util.hard_exit_from_instantiation(f"Failed to parse integration: {errors}")

        if not validation_errors:
            if entrypoint is not None:
                validation_errors += validate_entrypoint(entrypoint)
            if check_connections:
                validation_errors += validate_connections()

        validation_errors = [v.to_json() for v in validation_errors]
    except Exception:
        util.hard_exit_from_instantiation(
            message=f"Unexpected exception: {format_exc()}"
        )

    if validation_errors:
        exit_with_instantiation_errors(
            1, validation_errors, message="Integration validation failed."
        )
    else:
        print_successful_validation_result()


def validate_param(param):
    if not param.valid():
        return [ValidationError(message="Parameter must be defined", param=param)]
    try:
        param.read()
    except Exception as ex:
        return [ValidationError(message=str(ex), param=param)]
    return []


def validate_connections():
    errors = []
    try:
        for var in ConnectionParam.all():
            errors += validate_param(var)
        if not errors and connection_validator.validator is not None:
            return normalize_validation_result(connection_validator.validator())
    except Exception as ex:
        util.print_error(str(ex))
        exit(1)

    return errors


def normalize_validation_result(res):
    if res is None:
        return []
    if not isinstance(res, list):
        util.hard_exit_from_instantiation(
            f"Validator functions must return a list of ValidationError (got {res})"
        )
    else:
        return res


def validate_entrypoint(entrypoint):
    module_name = ".".join(entrypoint.split(".")[:-1])
    function_name = entrypoint.split(".")[-1]

    if module_name == "" or function_name == "":
        util.hard_exit_from_instantiation(
            "Bad entrypoint format. `Expected filename.functionname`"
        )

    # Try to import the world and find the entrypoint

    action_object = action.all().get(entrypoint)
    method = action_object.function
    execution_type = action_object.execution_type
    if method is None:
        util.hard_exit_from_instantiation(
            f"No matching action found. Is your action annotated with @action?"
        )
    if not isinstance(execution_type, ExecutionType):
        util.hard_exit_from_instantiation(
            f"Unsupported `execution_type` {str(execution_type)}: {type(execution_type)}. Valid execution-types are: {ExecutionType.get_all_types()}"
        )

    errors = []

    # Read the arguments and environment variables we expect and make sure they've all been defined
    parsed = parse(method.__doc__)
    args, _, _ = inspect.getargs(method.__code__)

    if execution_type == ExecutionType.PROCESS_ONCE and len(args) > 0:
        util.hard_exit_from_instantiation(
            f"`execution_type`: {str(execution_type)} should not have column-args. If you want to execute action once, use ROWS_TO_PROCESS input-fields"
        )

    for arg in args:
        optional = None
        for meta in parsed.meta:
            if len(meta.args) == 2 and meta.args[0] == "optional" and meta.args[1] == arg:
                optional = meta.description.lower() != "false"
        param = MappedColumnEnvVar(arg, optional=optional)
        if not param.valid():
            errors.append(
                ValidationError(
                    message="Column name cannot be empty", param=param
                )
            )

    env_vars = list(ConnectionParam.all()) + list(ActionParam.for_action(action_object))
    for var in env_vars:
        errors += validate_param(var)

    if errors:
        return errors

    if action_object.validator is not None:
        try:
            return normalize_validation_result(action_object.validator())
        except Exception as ex:
            util.hard_exit_from_instantiation(str(ex))
    return []


if __name__ == "__main__":
    main()



import inspect
import json
import re
from abc import ABCMeta
from collections import defaultdict
from typing import NamedTuple, Dict, Callable, Any

from lhub_integ import util
from lhub_integ.decorators import action as action_decorator
from lhub_integ.env import __EnvVar
from jinja2 import Template

from enum import Enum

ENV_VAR_NAME_REGEX = r"^[a-zA-Z_]\w*$"

TRUTHY = {"y", "true", "yes", "1"}
FALSY = {"n", "false", "no", "0"}


# pulled from forms.model.DataType
class DataType(Enum):
    STRING = "string"
    COLUMN = "column"
    NUMBER = "number"
    # no node datatype because integrations can only pull from one node

    # Unsupported by UI but validated by custom integrations
    BOOL = "bool"
    INT = "int"
    JSON = "json"
    JINJA = "jinja"

    def coerce(self, inp):
        if self == self.NUMBER:
            return float(inp)
        elif self == self.INT:
            return int(float(inp))
        elif self == self.BOOL:
            if isinstance(inp, bool):
                return inp
            if inp.lower() in TRUTHY:
                return True
            if inp.lower() in FALSY:
                return False
            raise ValueError(
                "Expected boolean input but {input} could not be coerced into a boolean value"
            )
        elif self == self.JSON:
            if isinstance(inp, dict):
                return inp
            return json.loads(inp)
        else:
            return inp


# pulled from forms.model.InputType
class InputType(Enum):
    TEXT = "text"
    TEXT_AREA = "textarea"
    EMAIL = "email"
    PASSWORD = "password"
    SELECT = "select"
    COLUMN_SELECT = "columnSelect"


class __Param(__EnvVar, metaclass=ABCMeta):
    def __init__(
            self,
            id,
            description=None,
            label=None,
            default=None,
            optional=False,
            options=None,
            data_type=DataType.STRING,
            input_type=InputType.TEXT,
    ):

        self.validate_id(id)
        super().__init__(id, default, optional)
        if label:
            self.label = label
        else:
            self.label = id
        self.description = description
        self.default = default
        self.data_type = data_type
        self.options = options
        if data_type == DataType.COLUMN:
            self.input_type = InputType.InputType.COLUMN_SELECT
        elif options is not None and len(options) > 1:
            self.input_type = InputType.SELECT
        else:
            self.input_type = input_type

    @staticmethod
    def validate_id(param_id: str):
        if re.match(ENV_VAR_NAME_REGEX, param_id) is None:
            util.invalid_integration(
                "invalid_parameter_name",
                f'"{param_id}" is not a valid id for a parameter. '
                f'Parameters must match `{ENV_VAR_NAME_REGEX}`. '
                'Use label to specify a custom display name',
            )

    def read(self):
        raw = super().read()
        return self.data_type.coerce(raw)


class JinjaTemplatedStr(str):
    @classmethod
    def input_type(cls) -> InputType:
        return InputType.TEXT_AREA

    @classmethod
    def data_type(cls) -> DataType:
        return DataType.JINJA


"""
We take most of the param information from our Form.Input case class. 
We don't enable a dependsOn field because if the dataType is a column then it will auto depends on its parent. 
"""


class ConnectionParam(__Param, metaclass=ABCMeta):
    """
    ConnectionParam provides a parameter specified by the connection

    Example usage:

    API_KEY = ConnectionParam('api_key')
    def process_row(url):
      requests.get(url, params={api_key: API_KEY.get()})
    """

    # LHUB-7385: Using simply a list here to preserve insertion order
    # list will suffice as we do check for duplicate connection-param id
    _all = []

    def __init__(self, *args, **kwargs):
        from lhub_integ import util

        super().__init__(*args, **kwargs)
        for conn_param in self._all:
            if conn_param.id == self.id:
                util.invalid_integration(
                    "duplication_connection_param",
                    f"You can't have 2 connection parameters with the same id ({self.id})",
                )
        self._all.append(self)

    @classmethod
    def all(cls):
        return cls._all


class ActionParam(__Param, metaclass=ABCMeta):
    """
    ActionParam provides a parameter specified by the action

    Example usage:

    API_KEY = ActionParam('api_key', action='process_row')

    import requests
    def process_row(url):
      requests.get(url, params={api_key: API_KEY.get()})
    """

    # LHUB-7385: defaultdict uses dict, so insertion order will be preserved. Also, changed from set to list to
    # preserve order of parameters. Since, we do have a duplicate check for action_parameters, this should be a safe
    # change.
    action_map = defaultdict(list)

    def __init__(self, *args, action, **kwargs):
        super().__init__(*args, **kwargs)
        from lhub_integ import util

        caller = inspect.currentframe().f_back
        if type(action) in (list, tuple):
            actions = action
        elif type(action) is str:
            actions = [action_str.strip() for action_str in action.split(',')]
        else:
            raise TypeError("'action' argument can be one of 'list', 'tuple' or 'str' only.")
        for action_str in actions:
            entrypoint = f"{caller.f_globals['__name__']}.{action_str}"
            if entrypoint in self.action_map:
                for action_param in self.action_map[entrypoint]:
                    if action_param.id == self.id:
                        util.invalid_integration(
                            "duplicate_action_param",
                            f"You can't have 2 action parameters with the same id ({action_str})",
                        )
            self.action_map[entrypoint].append(self)

    @classmethod
    def for_action(cls, action: action_decorator):
        return cls.action_map[action.entrypoint]


class ValidationError(NamedTuple):
    message: str
    param: "__EnvVar"

    def to_json(self):
        return {"message": self.message, "inputId": self.param.id}


CONVERTIBLE_TYPES = [int, str, float, bool, JinjaTemplatedStr]


def convert(c):
    def do_convert(raw: Dict[str, Any], column):
        if c == JinjaTemplatedStr:
            template = Template(column)
            return template.render(**raw)
        else:
            value = raw[column]
            if c in [str, float]:
                return c(value)
            if c == int:
                return int(float(value))
            elif c == bool:
                return DataType.BOOL.coerce(value)

    return do_convert


def get_input_converter(entrypoint_fn) -> Dict[str, Callable[[str], Any]]:
    """
    Build the input_conversion map to allow promotion from String to to int, float, and bool
    :param entrypoint_fn:
    :return: Dict from the name of the function arguments to a converter function.
    """
    sig = inspect.signature(entrypoint_fn)
    converter = {}
    from lhub_integ.util import exit_with_instantiation_errors

    for param in sig.parameters:
        annot = sig.parameters[param].annotation
        # The annotation is the Python class -- in these simple cases we can just call
        # the class constructor
        if annot in CONVERTIBLE_TYPES:
            converter[param] = convert(annot)
        elif annot == inspect.Parameter.empty:
            converter[param] = lambda raw, column: raw[column]
        else:
            exit_with_instantiation_errors(
                1,
                [
                    f"Unsupported type annotation: {annot}. Valid annotations are: {CONVERTIBLE_TYPES}"
                ],
            )
    return converter



from __future__ import print_function

import fileinput
import importlib
import inspect
import json
import sys
from pathlib import Path
from typing import List, Tuple


LHUB_ID = "lhub_id"


def read_all_data():
    """
    If an argument is provided, fileinput.input() will iterate over the lines in that file. Otherwise, it will
    read from stdin. See https://docs.python.org/3/library/fileinput.html
    :return:
    """
    return fileinput.input()


def hard_exit_from_instantiation(message):
    exit_with_instantiation_errors(2, [], message)


def exit_with_instantiation_errors(
    code, errors, message="Integration validation failed."
):
    error_wrapper = {"errors": errors, "message": message}
    print(f"[result] {json.dumps(error_wrapper)}", file=sys.stderr)
    sys.exit(code)


def invalid_integration(code, error):
    raise InvalidIntegration(code, error)


def print_error(message: str, data=None, original_lhub_id=None):
    error = {"has_error": True, "error": message, "data": data}
    print_result(json.dumps(error), original_lhub_id)


def print_result(msg, original_lhub_id=None):
    if original_lhub_id is not None:
        print_correlated_result(msg, original_lhub_id)
    else:
        print(f"[result] {msg}")


def print_correlated_result(msg, original_lhub_id):
    meta_data_dict = {"original_lhub_id": original_lhub_id}
    meta_json = json.dumps(meta_data_dict)
    print(f"[result][meta]{meta_json}[/meta] {msg}")


def print_successful_validation_result():
    print_result("{}")


def print_each_result_in_list(results, original_lhub_id=None):
    if not results:
        return print_result(
            json.dumps({"noResults": "no results returned"}),
            original_lhub_id=original_lhub_id,
        )
    for result in results:
        print_result(json.dumps(result), original_lhub_id=original_lhub_id)


def import_workdir() -> Tuple[List[Exception], List[str]]:
    """
    Attempt to import all the files in the working directory
    :return: A list of errors
    """
    errors = []
    docstrings = []
    for file in Path(".").iterdir():
        if file.suffix == ".py":
            as_module = file.name[: -len(".py")]
            if "." in as_module:
                errors.append(f"Python files cannot contain dots: {as_module}")
                continue
            try:
                module = importlib.import_module(as_module)
                doc = module.__doc__
                if doc:
                    docstrings.append(doc)
            except Exception as ex:
                errors.append(ex)
    return errors, docstrings


def import_file_of_entryPoint(entrypoint) -> Tuple[List[Exception], List[str]]:
    errors = []
    docstrings = []
    as_module = entrypoint.split(".")[0]
    try:
        module = importlib.import_module(as_module)
        doc = module.__doc__
        if doc:
            docstrings.append(doc)
    except Exception as ex:
        errors.append(ex)
    return errors, docstrings


def get_entrypoint_fn(entrypoint):
    module_name = ".".join(entrypoint.split(".")[:-1])
    function_name = entrypoint.split(".")[-1]
    module = importlib.import_module(module_name)
    return getattr(module, function_name)


def get_module_docstring(entrypoint):
    module_name = ".".join(entrypoint.split(".")[:-1])
    module = importlib.import_module(module_name)
    if module.__doc__:
        return module.__doc__.strip()
    return None


def type_annotations(fn):
    sig = inspect.signature(fn)
    return {param: sig.parameters[param].annotation for param in sig.parameters}


def deser_output(output: str):
    prefix = "[result]"
    processed = ""
    for line in output.splitlines():
        if line.startswith(prefix):
            processed += line[len(prefix) :]
    try:
        return json.loads(processed)
    except json.decoder.JSONDecodeError:
        raise Exception(f"Could not parse JSON: {output}")


class InvalidIntegration(Exception):
    def __init__(self, key, message):
        super().__init__(f"{message} [{key}]")



from enum import Enum


class ExecutionType(Enum):
    PROCESS_PER_ROW = "process_per_row"
    PROCESS_ONCE = "process_once"

    @staticmethod
    def get_all_types():
        return [exec_type.name for exec_type in ExecutionType]




import re
import time
import email.utils
import datetime
from io import StringIO
import csv

from bs4 import BeautifulSoup
from lhub_integ.common import helpers, time_helpers
from lhub_integ.common.input_helpers import safe_strip


def extract_urls(body):
    """
    Parse a plaintext email's message body, and return a sorted list of all unique URLs found

    :param body: plaintext message body
    :return: list of unique URLs
    """

    # In plaintext emails, sometimes URLs are simply placed in the body, but there is also a standard that some mail
    # services support, which is to enclose descriptive text in brackets and links in '<' and '>'
    # Example: [Google] <https://www.google.com/>

    # ToDo: Remove this old entry after vetting the new one
    # urls = re.findall(r'(?<=<)\w+:/{2,}(?:[\w\-]+\.)+\w+\S*?(?=>|\s|$)|(?<!<)\w+:/{2,}(?:[\w\-]+\.)+\w+\S*', body)

    # URL patterns
    url_base_pattern = r'\s*\w+:/{2,}(?:[\w\-]+\.)+\w+\S*?'
    url_patterns = [
        # Capture from body text when a URL is enclosed in '<' and '>' characters
        r'(?<=<){}(?=>|\s|$)'.format(url_base_pattern),

        # Capture from body text when a URL is enclosed in '[' and ']' characters
        r'(?<=\[){}(?=\]|\s|$)'.format(url_base_pattern),

        # Capture when URL appears directly within text but still contains 'xxxx://' prefix
        r'\b(?<!\S)\w+:/{2,}(?:[\w\-]+\.)+\w+\S*?(?=\"|<|\]|\s|$)',

        # Capture when "www.<domain>" appears within text, since Outlook will turn that into a hyperlink
        r'\s(www\.(?:[a-zA-Z0-9\-]+\.)+[a-zA-Z]+(?:/\S*)?)',

        # Capture when URL begins with an IP address and is followed by a slash
        r'\s((?:\d{1,3}\.){3}\d{1,3}/\S*)'
    ]

    # Look for URLs in the message body using all defined regex patterns
    urls = []
    for pattern in url_patterns:
        urls.extend([u.strip() for u in re.findall(pattern, body) if u.strip()])

    # If there are any empty values or mailto links, remove those
    urls_non_empty = []
    for u in urls:
        if u and not u.lower().startswith("mailto:") and not u.startswith('#'):
            urls_non_empty.append(u.replace('amp;', '').strip(',.'))

    # Remove duplicates and sort (for user readability)
    urls = sorted(list(set(urls_non_empty)))
    return urls


def extract_urls_from_html(html):
    """
    Parse an HTML email's message body, and return a sorted list of all unique URLs found
    :param html:
    :return: list of unique URLs
    """
    invalid_url_prefixes = ('mailto:', '#', 'tel:')

    def get_urls_from_tags(tags):
        extracted_urls = []
        for tag in tags:
            href = tag.get('href')
            # Sometimes there is an <a> tag without "href" but with a valid URL.
            # In that case, parse the link text w/ regex
            extracted_urls.append(href) if href else extracted_urls.extend(extract_urls(tag.text))
        return extracted_urls

    url_list = []
    soup = BeautifulSoup(html, 'lxml')
    links = soup.findAll('a')
    if links:
        urls = get_urls_from_tags(links)
        for url in urls:
            url = safe_strip(url)
            if url and not url.lower().startswith(invalid_url_prefixes):
                url_list.append(url.replace('amp;', ''))
        url_list = sorted(list(set(url_list)))

    return url_list


def __test_extract_urls_from_html():
    html = """<body><a href="https://logichub.com" /></body>"""
    urls = extract_urls_from_html(html)
    if urls[0] == 'https://logichub.com':
        helpers.print_debug_log("Extract URLs is working fine")
    else:
        helpers.print_debug_log("Extract URLs isn't working properly")


def default_email():
    log = helpers.format_success({})
    log.setdefault('lhub_ts', '%d' % (int(time.time()) * 1000))
    log.setdefault('sender', '')
    log.setdefault('recipients', [])
    log.setdefault('subject', '')
    log.setdefault('body', '')
    log.setdefault('body_text', '')
    log.setdefault('body_html', '')
    # ~Chad: new field to capture the body type (i.e. HTML vs. plaintext)
    log.setdefault('body_type', '')
    log.setdefault('attachments', [])
    log.setdefault('attachment_count', 0)
    log.setdefault('msgid', '')
    log.setdefault('date_received', '')
    log.setdefault('date_sent', '')
    log.setdefault('headers', [])
    log.setdefault('changekey', '')
    log.setdefault('categories', [])
    log.setdefault('urls', [])
    # ~Chad: now that we can pull unread messages too, adding a column for whether or not the message was unread
    log.setdefault('is_read', None)
    return log


def default_send_log(event_dict=None):
    event_dict = event_dict if event_dict else {}
    event_dict.update({
        'date_sent': time_helpers.current_time_string(),
        'recipients': '',
        'cc': '',
        'msg': '',
        'attachments': []
    })
    log = helpers.format_success(event_dict)
    return log


def add_recipients_to_list(recipient_list, existing_recipient_list):
    for mailbox in recipient_list:
        existing_recipient_list.append(mailbox.email_address)
    return existing_recipient_list


def add_recipients_to_list2(recipient_list, existing_recipient_list):
    if recipient_list:
        for mailbox in recipient_list:
            existing_recipient_list.append(mailbox['email'])
        return existing_recipient_list
    else:
        return existing_recipient_list


def parse_date_or_now(v):
    if v is None:
        return datetime.datetime.now()
    tt = email.utils.parsedate_tz(v)
    if tt is None:
        return datetime.datetime.now()
    timestamp = email.utils.mktime_tz(tt)
    date = datetime.datetime.fromtimestamp(timestamp)
    return date


email_re = re.compile(
    r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"  # dot-atom
    # quoted-string, see also http://tools.ietf.org/html/rfc2822#section-3.2.5
    r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"'
    r')@((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)$)'  # domain
    r'|\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$', re.IGNORECASE)

email_extract_re = re.compile(r"<(([.0-9a-z_+-=]+)@(([0-9a-z-]+\.)+[0-9a-z]{2,9}))>", re.M | re.S | re.I)


def extract_email(s):
    ret = email_extract_re.findall(s)
    if len(ret) < 1:
        p = s.split(" ")
        for e in p:
            e = e.strip()
            if email_re.match(e):
                return e

        return None
    else:
        return ret[0][0]


def parse_recipients(v):
    if v is None:
        return None

    ret = []

    # Sometimes a list is passed, which breaks .replace()
    if isinstance(v, list):
        v = ",".join(v)
    v = v.replace("\n", " ").replace("\r", " ").strip()
    s = StringIO(v)
    c = csv.reader(s)
    try:
        row = c.next()
    except StopIteration:
        return ret

    for entry in row:
        entry = entry.strip()
        if email_re.match(entry):
            e = entry
            entry = ""
        else:
            e = extract_email(entry)
            entry = entry.replace("<%s>" % e, "")
            entry = entry.strip()
            if e and entry.find(e) != -1:
                entry = entry.replace(e, "").strip()

        # If all else has failed
        if entry and e is None:
            e_split = entry.split(" ")
            e = e_split[-1].replace("<", "").replace(">", "")
            entry = " ".join(e_split[:-1])

        ret.append({"name": entry, "email": e})

    return ret



import os
import errno
import uuid
import base64
import gzip


class File:
    def __init__(self, dir, name, mode, auto_unzip=True):
        """
        Open, read, and provide various attributes of a file. Unless auto_unzip is disabled, will automatically unzip
        before reading if the file is found to be compressed in a supported file type.

        :param dir: Path to the file
        :param name: File name
        :param mode: Edit mode for the file (e.g. 'r' to read, 'w' to write, etc.)
        :param auto_unzip: Optional: set to False to disable automatic decompression
        """
        self.auto_unzip = auto_unzip
        self.__path = os.path.join(dir, name)
        self.name = name
        if "r" in mode and self.auto_unzip and get_mime_type(self.name) == 'application/gzip':
            self.fd = gzip.open(self.__path, mode)
        else:
            self.fd = open(self.__path, mode)

    def close(self):
        self.fd.close()

    def path(self):
        return self.__path

    def size(self):
        return os.path.getsize(self.__path)

    def base64_content(self):
        return base64.b64encode(self.fd.read())

    def mime_type(self):
        import magic
        return magic.from_file(self.__path, mime=True)

    def encoding(self):
        import chardet
        with open(self.__path, 'rb') as f:
            encoding = chardet.detect(f.read())
        return encoding

    def remove(self):
        os.remove(self.__path)


def get_mime_type(lhub_file_id):
    f = openFileForReading(lhub_file_id)
    t = f.mime_type()
    f.close()
    return t


def create_directory(path):
    try:
        os.makedirs(path)
    except OSError as ex:
        if ex.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def ensure_directory_exists(base_dir):
    if not os.path.isdir(base_dir):
        create_directory(base_dir)


def get_base_integration_directory():
    base_dir = os.getenv("INTEGRATIONS_FILES_PATH", "/tmp")
    ensure_directory_exists(base_dir)
    return base_dir


BASE_INTEGRATION_DIRECTORY = get_base_integration_directory()


def openNewFileForWriting(extension=None, mode="w"):
    name = uuid.uuid4().hex
    if extension:
        name += "." + extension
    return File(BASE_INTEGRATION_DIRECTORY, name, mode)


def openFileForReading(name, mode="r", auto_unzip=False):
    """
    Reads a file based on the provided name and returns a file object. Assumes that the read mode is "rU" ("readUniversal") by
    default, but the mode can be overridden if desired (e.g. "rb" to read as a binary file").
    Using mode 'rU' to open a file for reading in universal newline mode

    :param str name: lhub file ID
    :param str mode: Optional: mode to use for opening the file if other than "r" ("read")
    :param bool auto_unzip: Optional: if a file is found to be a supported zip type, unzip it before reading contents
    :return: Returns a file object
    """
    return File(BASE_INTEGRATION_DIRECTORY, name, mode, auto_unzip)



import json
import random
import string
import sys
import netaddr
import datetime
import dateutil
import dateutil.tz

import jinja2
import time
from collections import OrderedDict

from lhub_integ.common.input_helpers import __get_stripped_env_float
from lhub_integ.common import constants as global_constants


# Please make sure that Errors are passed in function in below format only:
# [{'message': 'blah-blah', 'inputId': 'access_key'}, .... , {'message': 'blah-again', 'inputId': 'server_url'}]
# otherwise your error message won't be able to be parsed correctly.


def print_validation_message(error):
    print("[result] %s" % json.dumps(error), file=sys.stderr)
    sys.exit(1)


def exit_with_instantiation_errors(code, errors, message="Integration validation failed."):
    error_wrapper = {"errors": errors, "message": message}
    print("[result] %s" % json.dumps(error_wrapper), file=sys.stderr)
    sys.exit(code)


def exit_with_error_message(code, msg):
    error = {"errors": [], "message": msg}
    print("[result] %s" % json.dumps(error), file=sys.stderr)
    sys.exit(code)


# Use this method when you do not care for fields from parent table
def print_result(result):
    if isinstance(result, dict):
        result = json.dumps(result)
    print("[result] %s" % result)


# Use this method when you do not care for fields from parent table
def print_error(msg):
    error = {"has_error": True, "error": msg}
    print("[result] %s" % json.dumps(error))


def format_success(event_dict=None):
    """
    Creates or updates an event dict to include the standard 'error' and 'has_error' fields with default values

    Args:
        event_dict (dict): Optional input to pass a dict of existing results to update. If none is provided, a dict
            containing only the new default fields will be returned.

    Returns:
        dict: New or updated dict containing default 'error' and 'has_error' keys
    """
    if not isinstance(event_dict, dict):
        event_dict = {"result": event_dict}
    event_dict.update({'error': None, 'has_error': False})
    return event_dict


def format_error(msg, event_dict=None):
    """
    Creates or updates an event dict to include the standard 'error' and 'has_error' fields for error results

    Args:
        msg (str): Error message string to include in the 'error' field of the output dict
        event_dict (dict): Optional input to pass a dict of existing results to update. If none is provided, a dict
            containing only the new default fields will be returned.

    Returns:
        dict: New or updated dict containing default 'error' and 'has_error' keys for error results

    """
    event_dict = event_dict if event_dict else {}
    event_dict.update({'error': msg, 'has_error': True})
    return event_dict


def print_correlated_error(msg, original_lhub_id, event_dict=None):
    """
    Prints an error result which will be correlated with the original lhub_id. If no dict value for event_dict is
    provided, the JSON event will contain only 'error' and 'has_error' fields.

    Args:
        msg (str): Error message (string) to place in the 'error' field in the JSON LogicHub output
        original_lhub_id (str or int): The lhub_id value from LogicHub source/parent table
        event_dict (dict): Optional dict of existing key-value pairs to include in the JSON LogicHub output

    Returns:
        N/A

    """
    event_dict = event_dict if event_dict else {}
    event_dict = format_error(msg, event_dict)
    meta_data_dict = {"original_lhub_id": original_lhub_id}
    print('[result][meta]{}[/meta]{}'.format(json.dumps(meta_data_dict), json.dumps(event_dict)))


# Function to use while serializing data structure using json.dumps if json contains datetime object.
def serialize_datetime_object(obj):
    if isinstance(obj, datetime.datetime):
        return obj.__str__()


# TODO ideally this would just be print_correlated_result(msg) with the system populating the original_lhub_id
def print_correlated_result(result, original_lhub_id):
    """
    :param result: could be a string or dict
    :param original_lhub_id: Correlation id
    :return:
    """
    meta_data_dict = {"original_lhub_id": original_lhub_id}
    if isinstance(result, dict):
        result = json.dumps(result)
    print('[result][meta]{}[/meta]{}'.format(json.dumps(meta_data_dict), result))


def print_debug_log(msg):
    try:
        print("[Debug] %s" % msg)
    except Exception as err:
        print("[Debug] Unknown exception while trying to print debug log: %s" % repr(err))


def print_successful_validation_result():
    print_result("{}")


def render_jinja_template(template_string, data, autoescape=False):
    if isinstance(template_string, str):
        unicode_template_string = template_string
    else:
        unicode_template_string = str(template_string, "utf-8")

    unicode_data = {}
    for key, value in list(data.items()):
        unicode_data[key] = value

    # undefined=jinja2.StrictUndefined this tells the parser to throw error if some variable is used in template but,
    # its value is not passed in the context environment
    jinja2_template = jinja2.Template(unicode_template_string, autoescape=autoescape, undefined=jinja2.StrictUndefined)
    return jinja2_template.render(unicode_data)


def convert_milliseconds_to_seconds(milliseconds):
    return milliseconds * 1.0 / 1000.0


def check_for_http_or_https_protocol(url):
    """ Checks for http:// or https:// protocol
        returns True if URL does NOT begin with http:// or https://
    """
    return url[:8] != "https://" and url[:7] != "http://"


def add_default_http_protocol_if_missing(url, protocol="http"):
    """ checks url for http protocol, adds a default http protocol if missing """
    if check_for_http_or_https_protocol(url):
        return protocol + "://" + url
    else:
        return url


def create_url_params_from_list(param_list):
    url_param_string = ''
    for url_param in param_list:
        if not url_param_string:
            url_param_string = '?'
        else:
            url_param_string = url_param_string + '&'
        url_param_string = url_param_string + str(url_param)
    return url_param_string


# Get a nested key from a dict, without having to do loads of ifs
def get_nested_key_value(results, keys):
    if isinstance(keys, list) and len(keys) > 0:

        if isinstance(results, dict) or isinstance(results, OrderedDict):
            key = keys.pop(0)
            if key in results:
                return get_nested_key_value(results[key], keys)
            else:
                return None
        else:
            if isinstance(results, list) and len(results) > 0:
                return get_nested_key_value(results[0], keys)
            else:
                return results
    else:
        return results


def print_each_result_in_list(results, original_lhub_id=None, return_result_in_json_format=True, encoder=str, ensure_ascii=True):
    if not results:
        if not original_lhub_id:
            return print_result(format_success({"noResults": "no results returned"}))
        else:
            return print_correlated_result(format_success({"noResults": "no results returned"}), original_lhub_id)
    for result in results:
        if return_result_in_json_format:
            result = json.dumps(format_success(result), default=encoder, sort_keys=True, ensure_ascii=ensure_ascii)
        if original_lhub_id is None:
            print_result(result)
        else:
            print_correlated_result(result, original_lhub_id)


def random_text(size):
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(size))


"""
@param size: length of random password desired
@param rules: array/dict of dicts. 
    Eg: [{"character_set": "1234567890", "size": 5}, {"character_set": "!@", "size": 5}]
    or
    {
        "alphabets": {"character_set": "abcdefghijklmnopqrstuvwxyz", "size": 5},
        "numbers": {"character_set": "1234567890", "size": 5},
        "special characters": {"character_set": "!@", "size": 5}
    }
@return: password string of length 'size'
"""
def password_generator(size, rules):
    password = ""

    if isinstance(rules, dict):
        rules = [rule for rule in rules.values()]
    for rule in rules:
        password += ''.join(random.choice(rule['character_set']) for _ in range(rule['size']))

    if len(password) > size:
        raise Exception("Invalid Password Rules")
    else:
        password_arr = list(password)
        for _ in range(len(password), size):
            password_arr.append(random.choice(password))
        random.shuffle(password_arr)
        return ''.join(password_arr)


def xstr(s):
    return None if s is None else str(s)


def string_or_none(s):
    if s:
        return s
    else:
        return None


def array_or_none(s):
    if s:
        try:
            return json.loads(s)
        except:
            return [s]
    else:
        return None


def int_of_string_or_none(s):
    if s:
        return int(s)
    else:
        return None


def keep_retrying(pull, check, sleep):
    r = pull()
    if check(r):
        return r
    else:
        time.sleep(sleep)
        return keep_retrying(pull, check, sleep)


def convert_xml_attr_string_in_json_to_json_property(json_obj, attr, custom_name=''):
    if not attr.startswith('@'):
        return json_obj
    new_key = custom_name if custom_name else attr[1:]
    json_obj[new_key] = json_obj[attr]
    del json_obj[attr]
    return json_obj


# helper function used in actions which are processed through Integration.process_multiple_rows_with_duplicate_columns
# Sometimes, we want a common msg/errorMsg for all rows in bucket supplied, so use this function in such cases.
def all_message_response_map(bucket, success, msg=""):
    result = {}
    for column_values in bucket:
        result[column_values["original_tuple"]] = success, msg
    return result


def get_remaining_time_for_action(start_time, timeout):
    time_elapsed = time.time() - start_time
    if time_elapsed < timeout:
        return timeout - time_elapsed
    else:
        return 0


def ip_address_as_string(ip_addr):
    try:
        return str(netaddr.IPAddress(ip_addr))
    except:
        return None


def ip_address_as_long(ip_addr):
    try:
        return int(netaddr.IPAddress(ip_addr))
    except:
        return None


def get_execution_time_range():
    start_datetime = __get_stripped_env_float(global_constants.EXECUTION_START_TIME) / 1000.0
    start_datetime = datetime.datetime.fromtimestamp(start_datetime, dateutil.tz.UTC)
    end_datetime = __get_stripped_env_float(global_constants.EXECUTION_END_TIME) / 1000.0
    end_datetime = datetime.datetime.fromtimestamp(end_datetime, dateutil.tz.UTC)
    return {"start_datetime": start_datetime, "end_datetime": end_datetime}



import os
import jinja2
import datetime

from lhub_integ.common import helpers
from lhub_integ.common.input_helpers import _get_stripped_env_string, _get_stripped_env_json, _get_stripped_env_yaml
from lhub_integ.common.input_helpers import _get_safe_stripped_env_integer, _get_json_from_string
from lhub_integ.common.query_parser import validate_and_parse, generate_query
import yaml.parser

# check_... methods return an error object if the given value is invalid
# validate_... methods additionally print the final result or exit


def create_error(msg, input_id=''):
    """
    Format an error response for validation of integration instances and actions

    Args:
        msg (str/unicode): Message to present to the user
        input_id (str): The environment variable name (what the variable is called in the descriptor and in inputs.py)

    Returns:
        (dict): Message and field ID to return to the user

    """
    return {"message": msg, "inputId": input_id}


def check_input_value_in_list_of_values(fieldname, input_id, list_of_values):
    value = _get_stripped_env_string(input_id)
    if not value:
        return create_error("{0} cannot be empty.".format(fieldname), input_id)
    elif value not in list_of_values:
        return create_error(
            "{0} is an invalid value. Must be one of: ({1}).".format(value, ', '.join(list_of_values)),
            input_id
        )
    else:
        return None


def check_optional_value_in_list_of_values(fieldname, input_id, list_of_values):
    value = _get_stripped_env_string(input_id)
    if not value:
        return None
    else:
        return check_input_value_in_list_of_values(fieldname, input_id, list_of_values)


def check_string(fieldname, value, input_id):
    """
    Check whether a given variable contains a value, and return a formatted error for integration input validation if it
    is empty.

    Args:
        fieldname (str): What it's called for the user
        value (str): The value entered by the user
        input_id (str): The environment variable name (what the variable is called in the descriptor and in inputs.py)

    Returns:
        None or a formatted validation error

    """
    if not value:
        return create_error("{0} cannot be empty.".format(fieldname), input_id)
    else:
        return None


def check_nonempty_string(input_id):
    value = _get_stripped_env_string(input_id)
    if not value:
        return create_error("Field cannot be empty.", input_id)
    else:
        return None


def check_nonempty_strings(*input_ids):
    for input_id in input_ids:
        value = _get_stripped_env_string(input_id)
        if not value:
            return create_error("Field cannot be empty.", input_id)
    return None


def check_non_negative_integer(input_id, value):
    try:
        intval = int(value)
        if intval < 0:
            return create_error("Invalid value ({0}), expected non-negative number".format(value), input_id)
        else:
            return None
    except ValueError:
        return create_error("Invalid value ({0}), expected integer".format(value), input_id)


def check_positive_integer(input_id, value):
    try:
        intval = int(value)
        if intval <= 0:
            return create_error("Invalid value ({0}), expected positive number".format(value), input_id)
        else:
            return None
    except ValueError:
        return create_error("Invalid value ({0}), expected integer".format(value), input_id)


def check_ranged_integer(input_id, value, minimum, maximum):
    error_msg = "Invalid value ({0}), expected a number between {1} and {2}".format(value, minimum, maximum)
    try:
        intval = int(value)
        if not(minimum <= intval <= maximum):
            return create_error(error_msg, input_id)
        else:
            return None
    except ValueError:
        return create_error(error_msg, input_id)


def check_optional_ranged_integer(input_id, minimum, maximum):
    value = _get_stripped_env_string(input_id)
    if value:
        return check_ranged_integer(input_id, value, minimum, maximum)
    else:
        return None


def check_non_negative_float(input_id, value):
    try:
        _var = float(value)
        if _var < 0:
            return create_error(
                "Invalid value ({0}), expected a non-negative number (whole or decimal)".format(value), input_id)
        else:
            return None
    except ValueError:
        return create_error(
            "Invalid value ({0}), expected a non-negative number (whole or decimal)".format(value), input_id)


def check_optional_non_negative_float(input_id):
    value = _get_stripped_env_string(input_id)
    if value:
        return check_non_negative_float(input_id, value)
    else:
        return None


def check_valid_json(*input_ids):
    try:
        for i in input_ids:
            value = _get_stripped_env_json(i)
        return None
    except ValueError:
        return create_error("Invalid value, expected json", i)


def check_valid_and_non_empty_yaml(*input_ids):
    try:
        for i in input_ids:
            value = _get_stripped_env_yaml(i)
            if not value:
                return create_error("Yaml shouldn't be empty", i)
        return None
    except yaml.parser.ParserError:
        return create_error("Invalid value, expected yaml", i)


def check_valid_templated_json(*input_ids):
    try:
        for input_id in input_ids:
            json_string = _get_stripped_env_string(input_id)
            validation_message, parsed_query_template_list = validate_and_parse(json_string)
            if validation_message:
                return create_error("Invalid value ({0})".format(validation_message), input_id)
            dummy_data = {}
            for template in parsed_query_template_list:
                if template[0] == 'VAR':
                    dummy_data[template[1]] = 123 # we are just passing a magic no to pass the validation
            query = generate_query(parsed_query_template_list, dummy_data)
            _get_json_from_string(query)
        return None
    except ValueError:
        return create_error("Invalid value ({0}), expected json".format(json_string), input_id)


def check_valid_templated_strings(*input_ids):
    try:
        for input_id in input_ids:
            json_string = _get_stripped_env_string(input_id)
            validation_message, parsed_query_template_list = validate_and_parse(json_string)
            if validation_message:
                return create_error("Invalid value ({0})".format(validation_message), input_id)
            dummy_data = {}
            for template in parsed_query_template_list:
                if template[0] == 'VAR':
                    dummy_data[template[1]] = 123 # we are just passing a magic no to pass the validation
            generate_query(parsed_query_template_list, dummy_data)
        return None
    except:
        return create_error("Invalid value ({0}), expected valid string".format(json_string), input_id)


def check_valid_template_string(input_id, accept_empty=False):
    template_content = _get_stripped_env_string(input_id)
    if not template_content and not accept_empty:
        return create_error("Template content cannot be empty.", input_id)

    if not template_content and accept_empty:
        return None
    try:
        if isinstance(template_content, str):
            unicode_template_content = template_content
        else:
            unicode_template_content = str(template_content, "utf-8")
        jinja2.Template(unicode_template_content)
    except jinja2.exceptions.TemplateSyntaxError as e:
        error_message = 'Syntax error on line {}: {}'.format(e.lineno, e.message)
        return create_error(error_message, input_id)
    else:
        return None


def check_allowed_options(input_id, field_name, options):
    value = _get_stripped_env_string(input_id)
    if not value:
        return create_error("Field '%s' cannot be empty" % field_name, input_id)
    elif value not in options:
        return create_error(
            "Field '%s' is invalid. Allowed options: %s" % (field_name, options), input_id)
    else:
        return None


def check_optional_non_negative_integer_field(input_id):
    value = _get_stripped_env_string(input_id)
    if value:
        return check_non_negative_integer(input_id, value)
    else:
        return None


def check_optional_date_time_format(input_id, format):
    value = _get_stripped_env_string(input_id)
    if value:
        try:
            datetime.datetime.strptime(value, format)
        except ValueError:
            return create_error("Invalid value ({0}) is not in expected format".format(value), input_id)
    else:
        return None


def check_url(input_id, value):
    from urllib.parse import urlparse
    try:
        urlparse(value)
    except:
        return create_error("Invalid value ({0}), expected URL".format(value), input_id)

    if helpers.check_for_http_or_https_protocol(value):
        return create_error("Invalid value ({0}), expected URL with http or https protocol".format(value), input_id)
    else:
        return None


def check_optional_url_field(input_id):
    value = _get_stripped_env_string(input_id)
    if value:
        return check_url(input_id, value)
    else:
        return None


def flatten(list_with_nones):
    if list_with_nones:
        return [_f for _f in list_with_nones if _f is not None]
    else:
        return list_with_nones


def check_column_name_and_delay(colname_var, timedelay_var):
    column_name = _get_stripped_env_string(colname_var)
    time_between_request = _get_stripped_env_string(timedelay_var)

    errors_with_nones = []

    errors_with_nones.append(check_string(
        "Column Name", column_name, colname_var))

    if time_between_request:
        errors_with_nones.append(check_non_negative_integer(
            timedelay_var, time_between_request))

    return flatten(errors_with_nones)


def check_column_name(colname_var):
    column_name = _get_stripped_env_string(colname_var)
   
    errors_with_nones = []

    errors_with_nones.append(check_string(
        "Column Name", column_name, colname_var))

    return flatten(errors_with_nones)


def check_column_names(*colnames_var):
    errors_with_nones = []
    for colname_var in colnames_var:
        column_name = _get_stripped_env_string(colname_var)
        errors_with_nones.append(check_string(
            "Column Name", column_name, colname_var))
    return flatten(errors_with_nones)


def check_delay_and_column_names(timedelay_var, *colnames_var):
    errors_with_nones = []
    time_between_request = _get_stripped_env_string(timedelay_var)
    if time_between_request:
        errors_with_nones.append(check_non_negative_integer(
            timedelay_var, time_between_request))
    for colname_var in colnames_var:
        column_name = _get_stripped_env_string(colname_var)
        errors_with_nones.append(check_string(
            "Column Name", column_name, colname_var))
    return flatten(errors_with_nones)


def validate_column_name_and_delay(colname_var, timedelay_var):
    finalize_validation_from_errors(
        check_column_name_and_delay(colname_var, timedelay_var)
    )


def validate_column_name(colname_var):
    finalize_validation_from_errors(
        check_column_name(colname_var)
    )


def validate_column_names(*colnames_var):
    finalize_validation_from_errors(
        check_column_names(*colnames_var)
    )


def validate_delay_and_column_names(timedelay_var, *colnames_var):
    finalize_validation_from_errors(
        check_delay_and_column_names(timedelay_var, *colnames_var)
    )


def validate_optional_api_key(api_key_var):
    api_key = os.environ.get(api_key_var)
    if api_key:
        validate_api_key(api_key_var)
    else:
        helpers.print_successful_validation_result()


def validate_api_key(api_key_var):
    api_key = _get_stripped_env_string(api_key_var)
    error = check_string("API Key", api_key, api_key_var)
    if error:
        helpers.exit_with_instantiation_errors(2, [error])
    else:
        helpers.print_successful_validation_result()


def check_string_params(label_varname_dict):
    errors_with_nones = []
    for label, env_var_name in list(label_varname_dict.items()):
        value = _get_stripped_env_string(env_var_name)
        errors_with_nones.append(check_string(label, value, env_var_name))
    return flatten(errors_with_nones)


def check_key_and_value(key_id, value_id):
    key = _get_stripped_env_string(key_id)
    value = _get_stripped_env_string(value_id)
    if bool(key) != bool(value):
        return create_error("Key and Value both should be entered.", key_id)
    else:
        return None


def validate_string_params(label_varname_dict):
    finalize_validation_from_errors(
        check_string_params(label_varname_dict)
    )


def check_username_password_exist(username_var, password_var):
    return check_string_params({
        "Username": username_var,
        "Password": password_var
    })


def finalize_validation_from_errors(errors):
    errors = flatten(errors)

    # Return results or exit
    if not errors:
        helpers.print_successful_validation_result()
    else:
        helpers.exit_with_instantiation_errors(2, errors)


def validate_username_password(username_var, password_var):
    errors = check_username_password_exist(username_var, password_var)
    finalize_validation_from_errors(errors)


def check_username_password_url(username_var, password_var, url_var):
    return check_string_params({
        "Username": username_var,
        "Password": password_var,
        "Instance URL": url_var,
    })


def check_server_port(server_var, port_var):
    errors = check_string_params({
        "Server": server_var,
        "Port": port_var
    })

    # validate server name
    server_value = _get_stripped_env_string(server_var)
    if ':' in server_value:
        errors.append(create_error("Server name cannot contain protocol", server_var))

    # validate port number, -1 indicates non-integer value given
    port_value = _get_safe_stripped_env_integer(port_var, -1)
    if port_value < 0 or port_value > 65535:
        errors.append(create_error("Port must be between 0 and 65535", port_var))

    return errors


def check_username_password_server_port(username_var, password_var, server_var, port_var):
    errors = check_string_params({
        "Username": username_var,
        "Password": password_var
    })

    if errors:
        return errors
    else:
        return check_server_port(server_var, port_var)


def check_username_apikey(username_var, apikey_var):
    return check_string_params({
        "Username": username_var,
        "API Key": apikey_var
    })


def check_action_timeout(timeout_var='__lh_action_timeout_sec'):
    return check_optional_non_negative_integer_field(timeout_var)


def validate_username_apikey(username_var, apikey_var):
    finalize_validation_from_errors(
        check_username_apikey(username_var, apikey_var)
    )


def validate_single_string(field_name, value, input_id=None):
    finalize_validation_from_errors(
        # a list of errors is expected, can include None
        [check_string(field_name, value, input_id)]
    )



import os
import tempfile
import subprocess

"""
Bundle all the dependencies in requirements.txt into a tar.bz2 by downloading and 
building wheels for all the dependencies in requirements.txt

See https://pip.pypa.io/en/stable/user_guide/#installation-bundles for details 
"""


def main():
    with tempfile.TemporaryDirectory() as tmpdir:
        if not os.path.exists("requirements.txt"):
            print("Run in a directory with requirements.txt")
            exit(1)
        subprocess.run(
            [
                "pip",
                "wheel",
                "-r",
                "requirements.txt",
                "--build-option",
                "--python-tag=py3",
                "--wheel-dir",
                tmpdir,
            ]
        )
        subprocess.run(["cp", "requirements.txt", tmpdir])
        cwd = os.getcwd()
        os.chdir(tmpdir)
        subprocess.run(["tar", "-cjvf", f"{cwd}/bundle.tar.bz2", *os.listdir(tmpdir)])
        os.chdir(cwd)



"""
`generate_metadata` will load a Python module and generate a metadata JSON that the module
system can use to build an integration descriptor.

Usage:
python -m lhub_integ.generate_metadata
"""
import inspect
import json
import re
from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json
from docstring_parser import parse

from lhub_integ import util, action
from lhub_integ.params import (
    ConnectionParam,
    ActionParam,
    DataType,
    InputType,
    JinjaTemplatedStr,
)
from lhub_integ.util import print_result, InvalidIntegration


@dataclass_json
@dataclass
class Param:
    id: str
    label: str
    data_type: str
    input_type: str
    optional: bool = False
    description: Optional[str] = None
    default: Optional[str] = None
    options: Optional[List[str]] = None


@dataclass_json
@dataclass
class Action:
    name: str
    entrypoint: str
    args: List[Param]
    params: List[Param]
    description: str
    errors: List[str]


@dataclass_json
@dataclass
class IntegrationMetadata:
    connection_params: List[Param]
    integration_name: str
    integration_description: str
    integration_logo_url: str
    actions: List[Action]
    errors: List[str]
    ok: bool


def generate_metadata() -> IntegrationMetadata:
    # Side-effectful action of importing every file in the current working directory.
    # This will populate the internal class-storage for `action`, `ConnectionParam` and `ActionParam`
    toplevel_errors = []
    parse_failure = False
    errors, docstrings = util.import_workdir()
    for error in errors:
        parse_failure = True
        if isinstance(error, SyntaxError):
            toplevel_errors.append(f"Failed to import module: {error}")
        elif isinstance(error, ImportError):
            toplevel_errors.append(
                f"Failed to import module (did you run bundle-integrations?): {error}"
            )
        elif isinstance(error, InvalidIntegration):
            toplevel_errors.append(f"Integration was invalid: {error}")
        else:
            toplevel_errors.append(
                f"Failed to import module (Unexpected error): {error}"
            )

    actions = []
    for action_object in action.all().values():
        entrypoint = action_object.entrypoint
        processing_function = action_object.function
        action_name = action_object.name

        # Parse the docstring
        docs = processing_function.__doc__
        parsed = parse(docs)

        errors = []

        # Any docstring on the function itself will be used as the description for the integration
        if parsed.short_description or parsed.long_description:
            function_description = parsed.short_description or parsed.long_description
        else:
            function_description = None

        # Build a map of the actual function arguments to compare with the docstrings
        args, varargs, kwargs = inspect.getargs(processing_function.__code__)
        type_annotations = util.type_annotations(processing_function)
        if varargs:
            errors.append("Varargs are not supported")
        if kwargs:
            errors.append("Kwargs are not supported")

        # default of a label is the id, may update it below when iterating over metadata
        arg_map = {
            arg: {"id": arg, "label": arg, "type": type_annotations[arg]}
            for arg in args
        }

        # load labels for column parameters
        # load default value for optional column parameters
        for meta in parsed.meta:
            if meta.args[0] == "label" and len(meta.args) == 2:
                arg_map[meta.args[1]]["label"] = meta.description
            elif len(meta.args) == 2 and meta.args[0] == "optional":
                arg_map[meta.args[1]]["optional"] = meta.description.lower() != "false"

        # Augment arg_map with information from the docstrings
        for param in parsed.params:
            if param.arg_name not in arg_map:
                errors.append(
                    f"Docstring referenced {param.arg_name} but there were no matching arguments"
                )
            else:
                if param.description is not None:
                    arg_map[param.arg_name]["description"] = param.description
                if param.type_name is not None:
                    arg_map[param.arg_name]["type"] = param.type_name

        params = [
            Param(
                id=e.id,
                label=e.label,
                description=e.description,
                default=e.default,
                optional=e.optional,
                data_type=e.data_type.value,
                options=e.options,
                input_type=e.input_type.value,
            )
            for e in ActionParam.for_action(action_object)
        ]
        args = [arg_map[arg] for arg in args]
        args = [
            Param(
                id=arg["id"],
                label=arg["label"],
                description=arg.get("description"),
                data_type=JinjaTemplatedStr.data_type().value
                if arg["type"] == JinjaTemplatedStr
                else DataType.COLUMN.value,
                input_type=JinjaTemplatedStr.input_type().value
                if arg["type"] == JinjaTemplatedStr
                else InputType.COLUMN_SELECT.value,
                optional=arg.get("optional", False)
            )
            for arg in args
        ]

        actions.append(
            Action(
                name=action_name,
                entrypoint=entrypoint,
                args=args,
                params=params,
                description=function_description,
                errors=errors,
            )
        )

    connection_params = [
        Param(
            id=e.id,
            label=e.label,
            description=e.description,
            default=e.default,
            optional=e.optional,
            data_type=e.data_type.value,
            options=e.options,
            input_type=e.input_type.value,
        )
        for e in ConnectionParam.all()
    ]

    if not actions and not parse_failure:
        toplevel_errors.append("No actions found. Did you forget to use @action?")

    integration_name = None
    integration_logo_url = None
    if docstrings:
        integration_description = get_info_from_docstrings(docstrings, "description")
        integration_name = get_info_from_docstrings(docstrings, "name")
        integration_logo_url = get_info_from_docstrings(docstrings, "logoUrl")
        if not integration_description:
            integration_description = docstrings[0].strip()

    elif len(actions) == 1:
        integration_description = actions[0].description
    else:
        integration_description = "No description provided"
    metadata = IntegrationMetadata(
        connection_params=connection_params,
        integration_name=integration_name,
        integration_description=integration_description,
        integration_logo_url=integration_logo_url,
        actions=actions,
        errors=toplevel_errors,
        ok=len(toplevel_errors) == 0,
    )
    return metadata


def get_info_from_docstrings(docstrings, info):
    for docstring in docstrings:
        split_string = re.split("{}:".format(info), docstring, flags=re.IGNORECASE)
        if len(split_string) >= 2:
            return split_string[1].splitlines()[0].strip()
    return None


if __name__ == "__main__":
    import traceback

    try:
        print_result(generate_metadata().to_json())
    except Exception:
        print_result(json.dumps(dict(errors=[traceback.format_exc()])))
        exit(1)



from lhub_integ.common import input_helpers

TYPE_LITERAL = 'LIT'
TYPE_VARIABLE = 'VAR'
DELIMITER_OPEN = '{{'
DELIMITER_CLOSE = '}}'

# Validates and Parses a Query Template string
# Ex:   'select {{col}} from {{tbl}} where {{cond}}'
#
# Returns a pair of (validation_message, parsed_query_template_list)
# Only one part of the pair will exist, the other will be None
#
# Example:
#
# (None,                       # validation message
#  [('LIT', 'select '),    # note whitespace
#   ('VAR', 'col'),
#   ('LIT', ' from '),
#   ('VAR', 'tbl'),
#   ('LIT', ' where '),
#   ('VAR', 'cond')])


def validate_and_parse(query_template_str):
    output = []
    parts = query_template_str.split(DELIMITER_OPEN)
    if len(parts) != len(query_template_str.split(DELIMITER_CLOSE)):
        return "Mismatch in number of open and close double-braces", None

    for part in parts:
        if DELIMITER_CLOSE in part:
            # must contain a variable then a literal
            subparts = part.split(DELIMITER_CLOSE)

            if len(subparts) != 2:
                return "Incorrect variable substitution near '{}' of '{}'".format(part, query_template_str), None

            variable_name = input_helpers.safe_strip(subparts[0])
            literal = subparts[1] # trailing whitespace is expected here

            if ' ' in variable_name:
                return "Variable cannot contain a space: '{}'".format(variable_name), None

            output.append(
                (TYPE_VARIABLE, variable_name)
            )

            if literal:
                output.append(
                    (TYPE_LITERAL, literal)
                )
        else:
            # must be a literal
            literal = part
            if literal:
                output.append(
                    (TYPE_LITERAL, literal)
                )
    return None, output


# Takes a parsed query template and a dict of variable names and values
# Fills in values of variables in the template and returns the query string
def generate_query(parsed_query_template_list, variable_name_value_dict):
    query_str = ''
    for element_type, element_value in parsed_query_template_list:
        if element_type == TYPE_LITERAL:
            query_str += element_value
        elif element_type == TYPE_VARIABLE:
            # Assumes column exists # TODO reconsider
            variable_name = element_value
            if variable_name in variable_name_value_dict:
                query_str += str(variable_name_value_dict.get(variable_name))
            else:
                raise ValueError("The following variable is missing from the input table: {}".format(variable_name))
        else:
            raise ValueError("Unknown query element type: {}".format(element_type))
    return query_str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lhub_integ', 'lhub_integ.common']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2',
 'PySocks==1.7.1',
 'PyYAML==5.3.1',
 'beautifulsoup4==4.8.2',
 'click>=7.0,<7.1',
 'dataclasses-json',
 'docstring-parser>=0.1,<0.2',
 'netaddr==0.7.19',
 'python-dateutil==2.8.1',
 'python-magic==0.4.15',
 'urllib3==1.25.8',
 'wheel>=0.32.3,<0.33.0']

entry_points = \
{'console_scripts': ['bundle-requirements = '
                     'lhub_integ.bundle_requirements:main']}

setup_kwargs = {
    'name': 'lhub-integ',
    'version': '1.0.9',
    'description': 'The Logichub Integration SDK',
    'long_description': "# lhub_integ\nPython package to shim basic scripts to work with integration machinery.\nThis package requires Python 3.6:\n```\n# Optional: install Python 3.6 with pyenvv\nbrew install pyenv\npyenv install 3.6.6\npyenv init # Follow the instructions\npyenv local 3.6.6\npython --version\n```\n\n```\npip install lhub_integ\n```\n\n## Usage (as an integration writer)\nTo write a Python script that is convertible into an integration:\n\n1. Create a directory that will contain your integration\n2. Install lhub_integ as a local package:\n```pip install lhub_integ```\n\nPython scripts must provide an entrypoint function with some number of arguments. These arguments will correspond to columns\nin the input data. The function should return a Python dictionary that can be serialized to JSON\n\n```python\ndef process(url, num_bytes: int):\n  return {'output': url + 'hello'}\n```\n\n### Specifying Dependencies\nYou must create a `requirements.txt` file specifying your dependencies. To create a dependency bundle run:\n```\n$ bundle-requirements\n```\nThis script is provided when you install lhub_integ.\n\n### Publishing\nYou will need the PyPi username and password that are in 1Password (search for PyPi)\n\n1. Bump the version in pyproject.toml\n\n2. `poetry publish --build`\n",
    'author': 'Russell Cohen',
    'author_email': 'russell@logichub.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://logichub.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import urllib3
import os

from lhub_integ.common.constants import *
from lhub_integ.common.input_helpers import _get_stripped_env_string


def verify_ssl_enabled(ssl_verify_default='True'):
    return _get_stripped_env_string(SSL_VERIFY_KEY, ssl_verify_default) != 'False'


if not verify_ssl_enabled():
    # Setting the path to CA cert empty, then it won't verify SSL certificate.
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    os.environ.__setitem__(PATH_TO_CA_CERT, '')




"""
    All Global Constants
"""

EXECUTION_START_TIME = '__execution_start_time_ms'
EXECUTION_END_TIME = '__execution_end_time_ms'

SSL_VERIFY_KEY = '__lh_ssl_verify'
THREADS_TO_USE = '__lh_multithread_count'
PATH_TO_CA_CERT = 'CURL_CA_BUNDLE'
TIME_BETWEEN_ROWS = "time_between_consecutive_requests_milliseconds"



import time
import dateutil.tz
from datetime import datetime


def current_time_int():
    return int(time.time())


def current_time_string(string_format="%Y-%m-%d %H:%M:%S %Z"):
    utc_now = datetime.now(tz=dateutil.tz.UTC)
    return utc_now.strftime(string_format)
